#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <dirent.h>
#include <pwd.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/wait.h>

#define MAX_BUF_LEN	256
#define APP_CHECKER_SYS_USER "linux-app-checker"

char app_checker_dir[PATH_MAX];
char app_data_dir[PATH_MAX];

void printUsage(const char* progname)
{
	printf("    Usage: %s (start|stop|restart) [port]\n", progname);
}

// Simulates `mkdir -p <path>`
// Return value: 1 if successful, 0 otherwise
int mkdir_p(const char* path)
{
	int pos = 1;
	char path_copy[PATH_MAX];
	struct stat stat_data;

	strcpy(path_copy, path);
	while (path_copy[pos] != '\0')
	{
		if (path_copy[pos] == '/')
		{
			path_copy[pos] = '\0';
			if (mkdir(path, 0755) == -1)
			{
				// Check if the path already exists
				if (errno == EEXIST)
				{
					// If it is not a directory, we have a problem
					if (stat(path, &stat_data) == -1)
						return 0;
					if (!S_ISDIR(stat_data.st_mode))
						return 0;
				}
			}
			path_copy[pos] = '/';
		}
		++pos;
	}
	return 1;
}

enum serverCommand
{
	START,
	STOP,
	RESTART
};

// Reads the information from the lock file specified, parses it and returns the specific data.
// Parameters:
//    lock_file [IN]        - path to the lock file to be read;
//    is_system [OUT]       - (optional) whether the corresponding server is started globally or not;
//    pid [OUT]             - PID of the running server;
//    port [OUT]            - (optional) port number occupied by the server;
//    username [OUT]        - (optional) name of the effective user of the server process;
//    username_buf_len [IN] - length of the input buffer for username.
// Return value:
//    1 if successful; 0 if the file could not be read or the server is not running.
int read_lock_file(const char* lock_file, int* is_system,
                   int* pid, int* port, char* username, size_t username_buf_len)
{
	char buf[MAX_BUF_LEN];
	char cmd_line[PATH_MAX];
	int fd, pos, res = 1;
	char* str_pos;
	FILE* stream;

	if ((fd = open(lock_file, O_RDONLY)) != -1)
	{
		if (flock(fd, LOCK_EX | LOCK_NB) == -1)
		{
			if ((pos = read(fd, buf, MAX_BUF_LEN - 1)) != -1)
			{
				buf[pos] = '\0';
				if (is_system != NULL)
					*is_system = ((buf[0] == 'G') ? 1 : 0);
				*pid = atoi(buf + ((((buf[0] == 'L') || (buf[0] == 'G')) && (buf[1] == ':')) ? 2 : 0));
				if ((str_pos = strstr(lock_file, "lock.")) != NULL)
				{
					if (port != NULL)
						*port = atoi(str_pos + 5);
					if (username != NULL)
					{
						// Try to get name of the owner of the process.
						// There seems to be no appropriate API call, the only way is to parse /proc/<PID>/status...
						// But calling `ps -o user= -p <PID>` is a bit simpler. :-)
						snprintf(cmd_line, PATH_MAX, "ps -o user= -p %d", *pid);
						if ((stream = popen(cmd_line, "r")) != NULL)
						{
							if ((pos = fread(buf, 1, MAX_BUF_LEN - 1, stream)) > 0)
							{
								buf[pos] = '\0';
								--pos;
								while ((pos > 0) && ((buf[pos] == '\r') || (buf[pos] == '\n')))
									buf[pos--] = '\0';
								if (pos + 2 > username_buf_len)
									buf[username_buf_len - 1] = '\0';
								strcpy(username, buf);
							}
							else
								res = 0;
							pclose(stream);
						}
						else
							res = 0;
					}
				}
				else
					res = 0;
			}
			else
				res = 0;
		}
		else
		{
			res = 0;
			flock(fd, LOCK_UN);
		}
		close(fd);
	}
	else
		res = 0;

	return res;
}

// Starts the App Checker server at the specified port and opens its page in a web-browser.
// If port_num equals to 0, the default 8889 will be used.
void serverStart(int port_num)
{
	char cmd_line[PATH_MAX];
	char lock_file[PATH_MAX];
	char buf[MAX_BUF_LEN];

	int i, pos, id, fd, browser_not_found, is_system;
	FILE* stream;
	
	char* known_browsers[] = {
		"opera",
		"/usr/bin/mozilla-firefox",
		"firefox",
		"iceweasel",
		"mozilla",
		"epiphany",
		"konqueror",
		"galeon",
		NULL
	};

	/* Try to check if another app-check-server already occupies the selected port */
	if (port_num == 0)
		port_num = 8889;
	snprintf(lock_file, PATH_MAX, "/tmp/app-checker-server.lock.%d", port_num);
	if (read_lock_file(lock_file, &is_system, &id, NULL, buf, MAX_BUF_LEN))
	{
		printf("    Error: An App Checker instance is running on the port '%d' already!\n    (Started by user '%s'; PID: %d)\n",
		       port_num, buf, id);
		exit(2);
	}

	if ((fd = open(lock_file, O_RDONLY)) != -1)
	{
		if (flock(fd, LOCK_EX | LOCK_NB) == -1)
		{
			printf("    Error: An App Checker instance is running on the port '%d' already!\n", port_num);
			if ((pos = read(fd, buf, MAX_BUF_LEN)) != -1)
			{
				buf[pos] = '\0';
				id = atoi(buf + ((((buf[0] == 'L') || (buf[0] == 'G')) && (buf[1] == ':')) ? 2 : 0));
				// Try to get name of the owner of that another instance.
				// There seems to be no appropriate API call, the only way is to parse /proc/<PID>/status...
				// But calling `ps -o user= -p <PID>` is a bit simpler. :-)
				snprintf(cmd_line, MAX_BUF_LEN, "ps -o user= -p %d", id);
				if ((stream = popen(cmd_line, "r")) != NULL)
				{
					if ((pos = fread(buf, 1, MAX_BUF_LEN, stream)) > 0)
					{
						buf[pos] = '\0';
						--pos;
						while ((pos > 0) && ((buf[pos] == '\r') || (buf[pos] == '\n')))
							buf[pos--] = '\0';
						printf("    (Started by user '%s'; PID: %d)\n", buf, id);
					}
					pclose(stream);
				}
			}
			exit(2);
		}
		else
		{
			flock(fd, LOCK_UN);
		}
		close(fd);
	}

	// Create necessary directories
	snprintf(cmd_line, PATH_MAX, "%s/results", app_data_dir);
	if (!mkdir_p(cmd_line))
	{
		printf("    Failed to create the results directory (%s)! (errno == %d)\n", cmd_line, errno);
		exit(2);
	}
	snprintf(cmd_line, PATH_MAX, "%s/log", app_data_dir);
	if (!mkdir_p(cmd_line))
	{
		printf("    Failed to create the log directory (%s)! (errno == %d)\n", cmd_line, errno);
		exit(2);
	}

	// Start the server in separate process
	id = fork();
	if (id == -1)
	{
		printf("    Failed to fork! (errno == %d)\n", errno);
		exit(2);
	}
	else if (id == 0)
	{
		// Child process: start App Checker server
		snprintf(cmd_line, PATH_MAX, "%s/webui/app-checker-server.pl", app_checker_dir);
		snprintf(buf, MAX_BUF_LEN, "%d", port_num);
		if (execl(cmd_line, "app-checker-server.pl", buf, NULL) == -1)
		{
			printf("    Failed to run app-checker-server.pl! (errno == %d)\n", errno);
			exit(2);
		}
	}

	// Check if the server did not fail
	sleep(1);
	if (waitpid(id, NULL, WNOHANG) > 0)
	{
		printf("    Failed to start server!\n");
		exit(2);
	}
	printf("    Server started. Log file location:\n    %s/log/app-checker-server.log.%d\n", app_data_dir, port_num);

	// Start the browser
	if (gethostname(buf, MAX_BUF_LEN) == -1)
		strcpy(buf, "localhost");
	else
		buf[MAX_BUF_LEN - 1] = '\0';
	printf("\n    The start page should be opened in a browser shortly.\n    If it doesn't open, you can load it at http://%s:%d\n", buf, port_num);
	
	browser_not_found = 5;

	// First, try the Portland's xdg-open util
	if (system("which xdg-open >/dev/null 2>&1") == 0)
	{
		// xdg-open exists
		snprintf(buf, MAX_BUF_LEN, "/tmp/appcheck-xdg-err.%d", getpid());
		unlink(buf);
		snprintf(cmd_line, PATH_MAX, "xdg-open \"http://127.0.0.1:%d\" >%s 2>&1", port_num, buf);
		browser_not_found = system(cmd_line);
		if (browser_not_found != 0)
		{
			// If unsuccessful, try to print the output of xdg-open
			if ((fd = open(buf, O_RDONLY)) != -1)
			{
				while ((pos = read(fd, cmd_line, PATH_MAX - 1)) != 0)
				{
					cmd_line[pos] = '\0';
					printf("%s", cmd_line);
				}
				close(fd);
			}
		}
		unlink(buf);
	}
	else
	{
		// No xdg-open was found
		for (i = 0; i < sizeof(known_browsers) / sizeof(known_browsers[0]); ++i)
		{
			snprintf(cmd_line, PATH_MAX, "which %s >/dev/null 2>&1", known_browsers[i]);
			if (system(cmd_line) == 0)
			{
				snprintf(cmd_line, PATH_MAX, "%s \"http://127.0.0.1:%d\" >/dev/null 2>&1 &", known_browsers[i], port_num);
				system(cmd_line);
				browser_not_found = 0;
				break;
			}
		}
	}
	if (browser_not_found)
	{
		printf("\n    Could not find a web-browser.\n");
	}
}

// Requests the user confirmation and terminates the program if reply is not 'yes'.
void confirm(const char* prompt) {
	char buf[MAX_BUF_LEN];
	int len;

	printf("%s", prompt);
	if (fgets(buf, MAX_BUF_LEN, stdin) == NULL)
		exit(3);
	len = strlen(buf);
	if ((buf[len - 1] == '\n') || (buf[len - 1] == '\r'))
		buf[len - 1] = '\0';
	if ((buf[len - 2] == '\n') || (buf[len - 2] == '\r'))
		buf[len - 2] = '\0';
	if ((strcasecmp(buf, "y") != 0) && (strcasecmp(buf, "yes") != 0))
		exit(3);
}

// Data for storing information about servers.
// Usernames (name_list) are stored for freeing allocated memory;
// port numbers - for quick navigation (rough replacement for hash arrays).
static char** name_list = NULL;
unsigned int* port_list = NULL;
static int list_num = 0;
static int list_sz = 0;

// Adds information to be stored.
int add_server_data(char* username, unsigned int port)
{
	if (username == NULL)
		return 1;
	if (list_num == list_sz)
	{
		if (list_sz == 0)
		{
			list_sz = 1024;
			name_list = malloc(list_sz * sizeof(char*));
			port_list = malloc(list_sz * sizeof(unsigned int));
		}
		else
		{
			list_sz *= 2;
			name_list = realloc(name_list, list_sz * sizeof(char*));
			port_list = realloc(port_list, list_sz * sizeof(unsigned int));
		}
		if ((name_list == NULL) || (port_list == NULL))
			return 0;
	}
	name_list[list_num] = username;
	port_list[list_num] = port;
	++list_num;
	return 1;
}

// Frees all the allocated memory.
void free_server_data()
{
	int i;

	if (name_list != NULL)
	{
		for (i = 0; i < list_num; ++i)
			free(name_list[i]);
		free(name_list);
	}
	if (port_list != NULL)
		free(port_list);
}

// unsigned int comparison function (for qsort).
int uint_cmp(const void* pa, const void* pb)
{
	unsigned int a = *(unsigned int*)pa, b = *(unsigned int*)pb;
	if (a < b)
		return -1;
	else if (a == b)
		return 0;
	else
		return 1;
}

// Stops the App Checker server running at the specified port, if any.
// If port_num equals to 0, and the only server is running, it will be stopped.
// If port_num equals to 0, and several servers are running, the user will be asked to select one.
// Only servers are listed that are started by the current user.
void serverStop(int port_num)
{
	char whoami[MAX_BUF_LEN];
	char buf[MAX_BUF_LEN];
	char cmd_line[PATH_MAX];
	struct passwd* pw_data;
	DIR* dd;
	struct dirent* dir_item;
	int srv_port, srv_pid, is_system, i, glob_i;
	uid_t euid = geteuid();
	char* tmp_buf;
	enum { ROOT, SYSTEM, NORMAL } role;

	char*        all_servers_usr[65536];   // port => user
	unsigned int all_servers[65536];       // port => pid
	unsigned int my_servers[65536];        // The same, but only for servers allowed to stop
	unsigned int my_servers_num = 0;

	memset(all_servers_usr, 0, 65536 * sizeof(char*));
	memset(all_servers, 0, 65536 * sizeof(char*));
	memset(my_servers, 0, 65536 * sizeof(char*));

	pw_data = getpwuid(euid);
	strncpy(whoami, pw_data->pw_name, MAX_BUF_LEN);

	// Determine the user permissions.
	// ROOT: the user is root => all servers can be stopped;
	// SYSTEM: the user started a suid binary from /opt/lsb/app-checker => can stop servers started from there;
	// NORMAL: the non-root user started a non-suid binary => can stop only non-suid servers started by him.
	if (euid == 0)
		role = ROOT;
	else if (strcmp(whoami, APP_CHECKER_SYS_USER) == 0)
		role = SYSTEM;
	else
		role = NORMAL;

	// Get the full list of running servers.
	if ((dd = opendir("/tmp")) == NULL)
	{
		printf("    Error: Could not access /tmp! (errno == %d)\n", errno);
		exit(2);
	}
	while ((dir_item = readdir(dd)) != NULL)
	{
		if ((strncmp(dir_item->d_name, "atk-server.lock.", 16) == 0) || (strncmp(dir_item->d_name, "app-checker-server.lock.", 24) == 0))
		{
			snprintf(cmd_line, PATH_MAX, "/tmp/%s", dir_item->d_name);
			if (read_lock_file(cmd_line, &is_system, &srv_pid, &srv_port, buf, MAX_BUF_LEN))
			{
				tmp_buf = strdup(buf);
				if (tmp_buf == NULL)
				{
					printf("    Error: malloc failed!\n");
					exit(3);
				}
				all_servers_usr[srv_port] = tmp_buf;
				if (!add_server_data(tmp_buf, srv_port))
				{
					printf("    Error: malloc failed!\n");
					exit(3);
				}
				all_servers[srv_port] = srv_pid;
				switch (role)
				{
				case ROOT:
					my_servers[srv_port] = srv_pid;
					++my_servers_num;
					break;
				case SYSTEM:
					if (is_system)
					{
						my_servers[srv_port] = srv_pid;
						++my_servers_num;
					}
					break;
				case NORMAL:
					if (strcmp(buf, whoami) == 0)
					{
						my_servers[srv_port] = srv_pid;
						++my_servers_num;
					}
					break;
				}
			}
		}
	}
	closedir(dd);
	qsort(port_list, list_num, sizeof(unsigned int), uint_cmp);

	if (port_num == 0)
	{
		// The users did not specify the port number.
		if (my_servers_num == 0)
		{
			// No suitable servers found.
			printf("    Error: No running App Checker servers %s were found.\n", ((list_num > 0) ? "started by you " : ""));
			exit(3);
		}
		else if (my_servers_num == 1)
		{
			// Only one suitable server found - select it.
			for (i = 0; i < list_num; ++i)
			{
				port_num = port_list[i];
				if (my_servers[port_num] != 0)
					break;
			}
			snprintf(buf, MAX_BUF_LEN, "    Are you sure you want to stop the server at port %d? ", port_num);
			confirm(buf);
		}
		else
		{
			// Several suitable servers found - ask the user to select one.
			printf("    There are several App Checker servers running:\n");
			glob_i = 0;
			// Pritn the formatted list of suitable servers.
			for (i = 0; i < list_num; ++i)
			{
				srv_port = port_list[i];
				if (my_servers[srv_port] != 0)
				{
					++glob_i;
					printf("    [%d]: Port %5d; PID: %d; owner: %s\n", glob_i, srv_port, my_servers[srv_port], all_servers_usr[srv_port]);
				}
			}
			printf("    Please specify the number of the server you want to stop: ");
			scanf("%d", &glob_i);
			if ((glob_i < 1) || (glob_i > my_servers_num))
			{
				printf("    Wrong number specified!\n");
				exit(3);
			}
			// Find the specific port number by the index entered
			for (i = 0; i < list_num; ++i)
			{
				port_num = port_list[i];
				if (my_servers[port_num] != 0)
				{
					--glob_i;
					if (glob_i == 0)
						break;
				}
			}
		}
	}
	else
	{
		// The user specified the port number.
		if (all_servers[port_num] == 0)
		{
			printf("    No App Checker server using the port '%d' was found.\n", port_num);
			exit(3);
		}
		if ((role != ROOT) && (my_servers[port_num] == 0))
		{
			printf("    The server was started by the user '%s' and you may not have permissions to stop it.\n", all_servers_usr[port_num]);
		}
		confirm("    Are you sure you want to stop the server? ");
	}

	free_server_data();

	snprintf(cmd_line, PATH_MAX, "%s/utils/stop_tests.pl %d", app_checker_dir, all_servers[port_num]);
	system(cmd_line);
}

int main(int argc, char* argv[])
{
	char* server_command_str;
	enum serverCommand server_command;
	int port_num = 0;
	int system_run;
	char buf[MAX_BUF_LEN];
	char exe_path[PATH_MAX];

	int pos, found;

	/* Check and remember input arguments */
	if ((argc < 2) || (argc > 3))
	{
		printUsage(argv[0]);
		exit(1);
	}
	server_command_str = argv[1];
	if (strcasecmp(server_command_str, "start") == 0)
		server_command = START;
	else if (strcasecmp(server_command_str, "stop") == 0)
		server_command = STOP;
	else if (strcasecmp(server_command_str, "restart") == 0)
		server_command = RESTART;
	else
	{
		printf("    Invalid command specified: '%s'.\n    Command should be one of the: start, stop, restart.\n", server_command_str);
		exit(1);
	}

	if (argc == 3)
	{
		port_num = atoi(argv[2]);
		snprintf(buf, MAX_BUF_LEN, "%d", port_num);
		if ((strcmp(buf, argv[2]) != 0) || (port_num < 1024) || (port_num > 65535))
		{
			printf("    Invalid port number specified: '%s'.\n    Port should be an integer number from range [1024, 65535].\n", argv[2]);
			exit(1);
		}
	}

	/* Determine path to ourselves */
	if (realpath(argv[0], exe_path) == NULL)
	{
		printf("    Failed to resolve executable path! (errno == %d)\n", errno);
		exit(2);
	}

	/* Extract path to app-checker dir */
	pos = strlen(exe_path) - 1;
	found = 0;
	while ((pos > 0) && (found < 2))
	{
		if (exe_path[pos] == '/')
			++found;
		--pos;
	}
	if (found < 2)
	{
		printf("    Failed to determine app-checker directory!\n");
		exit(2);
	}
	strncpy(app_checker_dir, exe_path, pos + 1);
	app_checker_dir[pos + 1] = '\0';
	if (strcmp(app_checker_dir, "/opt/lsb/app-checker") == 0)
	{
		system_run = 1;
		strcpy(app_data_dir, "/var/opt/lsb/app-checker");
	}
	else
	{
		system_run = 0;
		strcpy(app_data_dir, app_checker_dir);
	}

	if (server_command == START)
	{
		serverStart(port_num);
	}
	else if (server_command == STOP)
	{
		serverStop(port_num);
	}
	else if (server_command == RESTART)
	{
		printf("    Sorry, 'restart' command is not implemented yet.\n");
		exit(2);
	}
	return 0;
}
